import os
import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from datetime import datetime
from Bio.SeqRecord import SeqRecord


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


class MappingRecord:

    def __init__(self):

        #################### overall ####################

        self.qualified_reads = False

        #################### round 1 16s ####################

        self.consider_r1_unmapped_mate = False
        self.consider_r2_unmapped_mate = False

        self.r1_16s_ref_dict = dict()
        self.r2_16s_ref_dict = dict()

        self.r1_16s_refs_lowest_mismatch = None
        self.r2_16s_refs_lowest_mismatch = None

        self.r1_16s_refs_no_ignored = dict()
        self.r2_16s_refs_no_ignored = dict()
        self.shared_16s_refs_no_ignored = dict()

        self.both_mapped_to_16s = False

        #################### round 1 ctg ####################

        self.r1_ctg_ref_dict = dict()
        self.r2_ctg_ref_dict = dict()

        self.r1_ctg_refs_lowest_mismatch = None
        self.r2_ctg_refs_lowest_mismatch = None

        self.r1_ctg_refs_no_ignored = dict()
        self.r2_ctg_refs_no_ignored = dict()
        self.shared_ctg_refs_no_ignored = dict()

        self.matched_to_ctg = False

        #################### round 2 ####################

        self.qualified_reads_rd2 = False

        self.r1_ctg_ref_dict_rd2 = dict()
        self.r2_ctg_ref_dict_rd2 = dict()

        self.r1_ctg_refs_lowest_mismatch_rd2 = None
        self.r2_ctg_refs_lowest_mismatch_rd2 = None

        #################### round 2 mini_assembly ####################

        self.r1_mini_ref_dict = dict()
        self.r2_mini_ref_dict = dict()

        self.r1_mini_refs_lowest_mismatch = None
        self.r2_mini_refs_lowest_mismatch = None

        self.r1_mini_refs_no_ignored = dict()
        self.r2_mini_refs_no_ignored = dict()
        self.shared_mini_refs_no_ignored = dict()


def sam_flag_to_rc(flag_value):

    read_rced = 'na'
    if flag_value != '':
        binary_flag = "{0:b}".format(int(flag_value))
        binary_flag_len = len(str(binary_flag))
        binary_flag_polished = '0' * (12 - binary_flag_len) + str(binary_flag)

        if binary_flag_polished[7] == '0':
            read_rced = False
        if binary_flag_polished[7] == '1':
            read_rced = True

    return read_rced


def get_rc(seq_in):
    seq_in_rc = str(SeqRecord(Seq(seq_in)).reverse_complement().seq)
    return seq_in_rc


def get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len):

    mismatch_set_all_cigar = set()
    mismatch_set_long_M_cigars = set()
    for each_cigar in r1_ref_cigar_set:
        aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
        mismatch_set_all_cigar.add(mismatch_pct)
        if aligned_len >= min_M_len:
            mismatch_set_long_M_cigars.add(mismatch_pct)

    min_mismatch = 'NA'
    if len(mismatch_set_all_cigar) > 0:
        min_mismatch = min(mismatch_set_all_cigar)
        if len(mismatch_set_long_M_cigars) > 0:
            min_mismatch = min(mismatch_set_long_M_cigars)

    return min_mismatch


def get_max_clp_and_index(r1_cigar_list, r2_cigar_list):

    r1_cigar_list_split = [cigar_splitter(i) for i in r1_cigar_list]
    r2_cigar_list_split = [cigar_splitter(i) for i in r2_cigar_list]

    r1_cigar_list_split_only_clp = []
    for each_r1_cigar_split in r1_cigar_list_split:
        clp_len_l = 0
        if each_r1_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r1_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r1_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r1_cigar_split[-1][:-1])
        r1_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    r2_cigar_list_split_only_clp = []
    for each_r2_cigar_split in r2_cigar_list_split:
        clp_len_l = 0
        if each_r2_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r2_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r2_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r2_cigar_split[-1][:-1])
        r2_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    cigar_list_split_only_clp_r1_r2 = [r1_cigar_list_split_only_clp, r2_cigar_list_split_only_clp]

    max_value = 0
    max_value_index = ''
    for num_list_1 in cigar_list_split_only_clp_r1_r2[0]:
        if num_list_1[0] > max_value:
            max_value = num_list_1[0]
            max_value_index = 'r1_l'
        if num_list_1[1] > max_value:
            max_value = num_list_1[1]
            max_value_index = 'r1_r'
    for num_list_2 in cigar_list_split_only_clp_r1_r2[1]:
        if num_list_2[0] > max_value:
            max_value = num_list_2[0]
            max_value_index = 'r2_l'
        if num_list_2[1] > max_value:
            max_value = num_list_2[1]
            max_value_index = 'r2_r'

    # get the best cigar
    best_cigar = ''
    if max_value_index == 'r1_l':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r1_r':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_l':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_r':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar

    return best_cigar, max_value, max_value_index


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def check_cigar_quality(cigar_str, mismatch_cutoff, min_M_len, ref_pos, ref_len):
    r2_ctg_ref_cigar_splitted = cigar_splitter(cigar_str)
    qualified_cigar = False

    # check both end clip
    both_end_clp = check_both_ends_clipping(r2_ctg_ref_cigar_splitted)
    if both_end_clp is False:
        # check mismatch
        r2_aligned_len_ctg, r2_aligned_pct_ctg, r2_clipping_len_ctg, r2_clipping_pct_ctg, r2_mismatch_pct_ctg = get_cigar_stats(
            r2_ctg_ref_cigar_splitted)
        if r2_mismatch_pct_ctg <= mismatch_cutoff:
            # check aligned length
            if r2_aligned_len_ctg >= min_M_len:
                # check if clp in the middle
                clip_in_middle = False
                if ('S' in cigar_str) or ('s' in cigar_str):
                    clip_in_middle = True
                    if (r2_ctg_ref_cigar_splitted[0][-1] in ['S', 's']) and (ref_pos == 1):
                        clip_in_middle = False
                    if (r2_ctg_ref_cigar_splitted[-1][-1] in ['S', 's']):
                        if (ref_pos + r2_aligned_len_ctg - 1) == ref_len:
                            clip_in_middle = False

                if clip_in_middle is False:
                    qualified_cigar = True

    return qualified_cigar


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '..'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, within_gnm_linkage_num_diff, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:
                if MarkerGene not in MarkerGene_with_assignment:

                    # consider depth
                    if min_16s_gnm_multiple > 0:

                        # get marker and genome depth
                        MarkerGene_depth = marker_gene_depth_dict.get(MarkerGene, 'na')
                        GenomicSeq_depth = genomic_seq_depth_dict.get(GenomicSeq, 'na')
                        marker_genome_depth_ratio = 'na'
                        if (MarkerGene_depth != 'na') and (GenomicSeq_depth != 'na'):
                            if GenomicSeq_depth > 0:
                                marker_genome_depth_ratio = MarkerGene_depth / GenomicSeq_depth

                        if marker_genome_depth_ratio >= min_16s_gnm_multiple:
                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                gnm_max_link_num_dict[GenomicSeq] = linkage_num
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                # get identity with best marker
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                                iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                    if (linkage_num * 100 / gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                        file_out_handle.write(each_match)
                                        MarkerGene_with_assignment.add(MarkerGene)
                                    else:
                                        MarkerGene_with_assignment.add(MarkerGene)
                    # ignore depth
                    else:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            gnm_max_link_num_dict[GenomicSeq] = linkage_num
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            # get identity with best marker
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                            iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                if (linkage_num*100/gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                                else:
                                    MarkerGene_with_assignment.add(MarkerGene)
    file_out_handle.close()


##################################################### file in ####################################################

step_1_wd = '/Users/songweizhi/Desktop/tunning_rd1'
input_16s_polished                          = '%s/file_in/MBARC26_SILVA138_polished.fasta'                          % step_1_wd
combined_input_gnms                         = '%s/file_in/Refined_refined_bins_renamed_combined.fa'                 % step_1_wd
input_reads_to_16s_sam_sorted               = '%s/file_in/MBARC26_0521_input_reads_to_16S_reformatted_sorted.sam'   % step_1_wd
input_reads_to_16s_sam_sorted               = '%s/file_in/MBARC26_0521_input_reads_to_16S_reformatted_sorted_subset.sam'   % step_1_wd
blast_results_all_vs_all_16s                = '%s/file_in/MBARC26_0521_16S_all_vs_all_blastn.tab'                   % step_1_wd
rd1_extracted_to_gnm_sam_reformatted_sorted = '%s/file_in/rd1_extracted_to_gnm_reformatted_sorted.sam'              % step_1_wd
min_M_len_16s                               = 45
min_M_len_ctg                               = 45
min_M_pct                                   = 35
mismatch_cutoff                             = 2
read_to_marker_connector                    = '___r___'
marker_to_ctg_gnm_Key_connector             = '___M___'
gnm_to_ctg_connector                        = '___C___'
min_aln_16s = 500
min_cov_16s = 30
mean_depth_dict_gnm = {}
mean_depth_dict_16s = {}
min_16s_gnm_multiple = 0
min_iden_16s = 98
min_link_num = 8
within_gnm_linkage_num_diff = 80
time_format = '[%Y-%m-%d %H:%M:%S]'

rd1_clp_pct_diff_txt                        = '%s/rd1_clp_pct_diff.txt'             % step_1_wd
rd1_clp_pct_diff_txt_to_ignore              = '%s/rd1_clp_pct_diff_to_ignore.txt'   % step_1_wd
link_stats_combined                         = '%s/stats_combined.txt'               % step_1_wd
link_stats_combined_filtered_s1             = '%s/stats_combined_filtered.txt'      % step_1_wd
linking_reads_rd1                           = '%s/linking_reads_rd1.txt'            % step_1_wd


##################################################### read in sam file ####################################################

print('%s %s' % ((datetime.now().strftime(time_format)), 'read in sam file'))

# get marker len dict
marker_len_dict = {}
for each_marker_record in SeqIO.parse(input_16s_polished, 'fasta'):
    marker_len_dict[each_marker_record.id] = len(each_marker_record.seq)

processed_num = 0
MappingRecord_dict = {}
to_extract_read_base = set()
current_read_base = ''
current_read_base_r1_16s_ref_dict = dict()
current_read_base_r2_16s_ref_dict = dict()
with open(input_reads_to_16s_sam_sorted) as input_reads_to_16s_sam_opened:
    for each_read in input_reads_to_16s_sam_opened:
        if not each_read.startswith('@'):
            each_read_split = each_read.strip().split('\t')
            cigar = each_read_split[5]
            read_id = each_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])

            if current_read_base == '':
                current_read_base = read_id_base

                if cigar != '*':
                    if read_strand == '1':
                        current_read_base_r1_16s_ref_dict[ref_id] = {ref_pos: cigar}
                    if read_strand == '2':
                        current_read_base_r2_16s_ref_dict[ref_id] = {ref_pos: cigar}

            elif read_id_base == current_read_base:

                if cigar != '*':
                    if read_strand == '1':
                        if ref_id not in current_read_base_r1_16s_ref_dict:
                            current_read_base_r1_16s_ref_dict[ref_id] = {ref_pos: cigar}
                        else:
                            current_read_base_r1_16s_ref_dict[ref_id][ref_pos] = cigar
                    if read_strand == '2':
                        if ref_id not in current_read_base_r2_16s_ref_dict:
                            current_read_base_r2_16s_ref_dict[ref_id] = {ref_pos: cigar}
                        else:
                            current_read_base_r2_16s_ref_dict[ref_id][ref_pos] = cigar
            else:
                ################################### analysis previous read refs ####################################

                ########## get lowest mismatch for r1/r2 16s refs ##########

                # get r1_ref_cigar_set
                r1_ref_cigar_set = set()
                for each_pos_dict in current_read_base_r1_16s_ref_dict.values():
                    each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
                    r1_ref_cigar_set.update(each_pos_dict_values)

                # get r2_ref_cigar_set
                r2_ref_cigar_set = set()
                for each_pos_dict in current_read_base_r2_16s_ref_dict.values():
                    each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
                    r2_ref_cigar_set.update(each_pos_dict_values)

                r1_ref_min_mismatch = get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len_16s)
                r2_ref_min_mismatch = get_min_mismatch_from_cigar_list(r2_ref_cigar_set, min_M_len_16s)

                refs_to_ignore = set()

                ########## filter r1 16s refs ##########

                r1_16s_refs_passed_qc = {}
                for r1_16s_ref in current_read_base_r1_16s_ref_dict:
                    r1_matched_pos_dict = current_read_base_r1_16s_ref_dict[r1_16s_ref]

                    # one read need to mapped to one 16S only for one time
                    if len(r1_matched_pos_dict) > 1:
                        refs_to_ignore.add(r1_16s_ref)
                    else:
                        r1_16s_ref_pos = list(r1_matched_pos_dict.keys())[0]
                        r1_16s_ref_cigar = r1_matched_pos_dict[r1_16s_ref_pos]
                        r1_16s_ref_cigar_splitted = cigar_splitter(r1_16s_ref_cigar)

                        # check both end clip
                        both_end_clp = check_both_ends_clipping(r1_16s_ref_cigar_splitted)
                        if both_end_clp is True:
                            refs_to_ignore.add(r1_16s_ref)
                        else:
                            # check mismatch
                            r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(
                                r1_16s_ref_cigar_splitted)
                            if r1_ref_min_mismatch == 'NA':
                                refs_to_ignore.add(r1_16s_ref)
                            elif (r1_mismatch_pct > r1_ref_min_mismatch) or (r1_mismatch_pct > mismatch_cutoff):
                                refs_to_ignore.add(r1_16s_ref)
                            else:
                                # check aligned length
                                if r1_aligned_len < min_M_len_16s:
                                    refs_to_ignore.add(r1_16s_ref)
                                else:
                                    # check if clp in the middle
                                    clip_in_middle = False
                                    if ('S' in r1_16s_ref_cigar) or ('s' in r1_16s_ref_cigar):
                                        clip_in_middle = True
                                        if (r1_16s_ref_cigar_splitted[0][-1] in ['S', 's']) and (
                                                r1_16s_ref_pos == 1):
                                            clip_in_middle = False
                                        if (r1_16s_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                            if (r1_16s_ref_pos + r1_aligned_len - 1) == marker_len_dict[r1_16s_ref]:
                                                clip_in_middle = False

                                    # exclude the ref if clp in the middle is True
                                    if clip_in_middle is True:
                                        refs_to_ignore.add(r1_16s_ref)
                                    else:
                                        r1_16s_refs_passed_qc[r1_16s_ref] = [r1_16s_ref_cigar]

                ########## filter r2 16s refs ##########

                r2_16s_refs_passed_qc = {}
                for r2_16s_ref in current_read_base_r2_16s_ref_dict:
                    r2_matched_pos_dict = current_read_base_r2_16s_ref_dict[r2_16s_ref]

                    # one read need to mapped to one 16S only once
                    if len(r2_matched_pos_dict) > 1:
                        refs_to_ignore.add(r2_16s_ref)
                    else:
                        r2_16s_ref_pos = list(r2_matched_pos_dict.keys())[0]
                        r2_16s_ref_cigar = r2_matched_pos_dict[r2_16s_ref_pos]
                        r2_16s_ref_cigar_splitted = cigar_splitter(r2_16s_ref_cigar)

                        # check both end clip
                        both_end_clp = check_both_ends_clipping(r2_16s_ref_cigar_splitted)
                        if both_end_clp is True:
                            refs_to_ignore.add(r2_16s_ref)
                        else:
                            # check mismatch
                            r2_aligned_len, r2_aligned_pct, r2_clipping_len, r2_clipping_pct, r2_mismatch_pct = get_cigar_stats(
                                r2_16s_ref_cigar_splitted)
                            if r2_ref_min_mismatch == 'NA':
                                refs_to_ignore.add(r2_16s_ref)
                            elif (r2_mismatch_pct > r2_ref_min_mismatch) or (r2_mismatch_pct > mismatch_cutoff):
                                refs_to_ignore.add(r2_16s_ref)
                            else:
                                # check aligned length
                                if r2_aligned_len < min_M_len_16s:
                                    refs_to_ignore.add(r2_16s_ref)
                                else:
                                    # check if clp in the middle
                                    clip_in_middle = False
                                    if ('S' in r2_16s_ref_cigar) or ('s' in r2_16s_ref_cigar):
                                        clip_in_middle = True
                                        if (r2_16s_ref_cigar_splitted[0][-1] in ['S', 's']) and (
                                                r2_16s_ref_pos == 1):
                                            clip_in_middle = False
                                        if (r2_16s_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                            if (r2_16s_ref_pos + r2_aligned_len - 1) == marker_len_dict[r2_16s_ref]:
                                                clip_in_middle = False

                                    # exclude the ref if clp in the middle is True
                                    if clip_in_middle is True:
                                        refs_to_ignore.add(r2_16s_ref)
                                    else:
                                        r2_16s_refs_passed_qc[r2_16s_ref] = [r2_16s_ref_cigar]

                #################################

                r1_16s_refs_no_ignored = {key: value for key, value in r1_16s_refs_passed_qc.items() if
                                          key not in refs_to_ignore}
                r2_16s_refs_no_ignored = {key: value for key, value in r2_16s_refs_passed_qc.items() if
                                          key not in refs_to_ignore}

                # no mate has no_ignored alignments
                if (len(r1_16s_refs_no_ignored) == 0) and (len(r2_16s_refs_no_ignored) == 0):
                    pass

                # only r1 has no_ignored alignments
                elif (len(r1_16s_refs_no_ignored) > 0) and (len(r2_16s_refs_no_ignored) == 0):

                    if current_read_base not in MappingRecord_dict:
                        MappingRecord_dict[current_read_base] = MappingRecord()

                    MappingRecord_dict[current_read_base].qualified_reads = True
                    MappingRecord_dict[current_read_base].consider_r1_unmapped_mate = True
                    MappingRecord_dict[current_read_base].r1_16s_refs_no_ignored = r1_16s_refs_no_ignored
                    MappingRecord_dict[current_read_base].r1_16s_ref_dict = current_read_base_r1_16s_ref_dict
                    to_extract_read_base.add(current_read_base)

                # only r2 has no_ignored alignments
                elif (len(r1_16s_refs_no_ignored) == 0) and (len(r2_16s_refs_no_ignored) > 0):

                    if current_read_base not in MappingRecord_dict:
                        MappingRecord_dict[current_read_base] = MappingRecord()

                    MappingRecord_dict[current_read_base].qualified_reads = True
                    MappingRecord_dict[current_read_base].consider_r2_unmapped_mate = True
                    MappingRecord_dict[current_read_base].r2_16s_refs_no_ignored = r2_16s_refs_no_ignored
                    MappingRecord_dict[current_read_base].r2_16s_ref_dict = current_read_base_r2_16s_ref_dict

                    to_extract_read_base.add(current_read_base)

                # both r1 and r2 have no_ignored alignments
                else:
                    shared_16s_ref_dict = {key: [r1_16s_refs_no_ignored[key][0], r2_16s_refs_no_ignored[key][0]] for key
                                           in set(r1_16s_refs_no_ignored).intersection(set(r2_16s_refs_no_ignored))}
                    if len(shared_16s_ref_dict) > 0:

                        if current_read_base not in MappingRecord_dict:
                            MappingRecord_dict[current_read_base] = MappingRecord()

                        MappingRecord_dict[current_read_base].qualified_reads = True
                        MappingRecord_dict[current_read_base].both_mapped_to_16s = True
                        MappingRecord_dict[current_read_base].shared_16s_refs_no_ignored = shared_16s_ref_dict
                        MappingRecord_dict[current_read_base].r1_16s_ref_dict = current_read_base_r1_16s_ref_dict
                        MappingRecord_dict[current_read_base].r2_16s_ref_dict = current_read_base_r2_16s_ref_dict
                        to_extract_read_base.add(current_read_base)

                    '''
                    qualified_reads
                    consider_r1_unmapped_mate
                    consider_r2_unmapped_mate
                    both_mapped_to_16s
                    r1_16s_ref_dict
                    r2_16s_ref_dict
                    r1_16s_refs_no_ignored
                    r2_16s_refs_no_ignored
                    shared_16s_refs_no_ignored
                    '''
                # if current_read_base in MappingRecord_dict:
                #     print('%s\tr1_16s_ref_dict\t%s'             % (current_read_base, MappingRecord_dict[current_read_base].r1_16s_ref_dict))
                #     print('%s\tr2_16s_ref_dict\t%s'             % (current_read_base, MappingRecord_dict[current_read_base].r2_16s_ref_dict))
                #     print('%s\tr1_16s_refs_no_ignored\t%s'      % (current_read_base, MappingRecord_dict[current_read_base].r1_16s_refs_no_ignored))
                #     print('%s\tr2_16s_refs_no_ignored\t%s'      % (current_read_base, MappingRecord_dict[current_read_base].r2_16s_refs_no_ignored))
                #     print('%s\tshared_16s_refs_no_ignored\t%s'  % (current_read_base, MappingRecord_dict[current_read_base].shared_16s_refs_no_ignored))
                #     print()
                #
                #     print()


                ########################################### reset values ###########################################

                current_read_base = read_id_base
                current_read_base_r1_16s_ref_dict = dict()
                current_read_base_r2_16s_ref_dict = dict()

                if cigar != '*':
                    if read_strand == '1':
                        current_read_base_r1_16s_ref_dict[ref_id] = {ref_pos: cigar}
                    if read_strand == '2':
                        current_read_base_r2_16s_ref_dict[ref_id] = {ref_pos: cigar}

current_read_base = 'MBARC26_251146'
print('%s\tr1_16s_ref_dict\t%s'             % (current_read_base, MappingRecord_dict[current_read_base].r1_16s_ref_dict))
print('%s\tr2_16s_ref_dict\t%s'             % (current_read_base, MappingRecord_dict[current_read_base].r2_16s_ref_dict))
print('%s\tr1_16s_refs_no_ignored\t%s'      % (current_read_base, MappingRecord_dict[current_read_base].r1_16s_refs_no_ignored))
print('%s\tr2_16s_refs_no_ignored\t%s'      % (current_read_base, MappingRecord_dict[current_read_base].r2_16s_refs_no_ignored))
print('%s\tshared_16s_refs_no_ignored\t%s'  % (current_read_base, MappingRecord_dict[current_read_base].shared_16s_refs_no_ignored))
print()


######################################### read mapping results of rd1 extracted mates into mp dict  #########################################

print('%s %s' % ((datetime.now().strftime(time_format)), 'read in sam file'))

ctg_len_dict = {}
for each_ctg_record in SeqIO.parse(combined_input_gnms, 'fasta'):
    ctg_len_dict[each_ctg_record.id] = len(each_ctg_record.seq)


current_read_base = ''
current_read_base_r1_ctg_ref_dict = dict()
current_read_base_r2_ctg_ref_dict = dict()
with open(rd1_extracted_to_gnm_sam_reformatted_sorted) as rd1_extracted_to_gnm_sam_reformatted_opened:
    for each_read in rd1_extracted_to_gnm_sam_reformatted_opened:
        if not each_read.startswith('@'):
            each_read_split = each_read.strip().split('\t')
            cigar = each_read_split[5]
            read_id = each_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])

            if current_read_base == '':
                current_read_base = read_id_base
                if cigar != '*':
                    if read_strand == '1':
                        current_read_base_r1_ctg_ref_dict[ref_id] = {ref_pos: cigar}
                    if read_strand == '2':
                        current_read_base_r2_ctg_ref_dict[ref_id] = {ref_pos: cigar}

            elif read_id_base == current_read_base:
                if cigar != '*':
                    if read_strand == '1':
                        if ref_id not in current_read_base_r1_ctg_ref_dict:
                            current_read_base_r1_ctg_ref_dict[ref_id] = {ref_pos: cigar}
                        else:
                            current_read_base_r1_ctg_ref_dict[ref_id][ref_pos] = cigar

                    if read_strand == '2':
                        if ref_id not in current_read_base_r2_ctg_ref_dict:
                            current_read_base_r2_ctg_ref_dict[ref_id] = {ref_pos: cigar}
                        else:
                            current_read_base_r2_ctg_ref_dict[ref_id][ref_pos] = cigar
            else:
                ################################### analysis previous read refs ####################################

                refs_to_ignore_ctg = set()

                ########## filter r1 ctg refs ##########

                r1_ctg_refs_passed_qc = {}
                for r1_ctg_ref in current_read_base_r1_ctg_ref_dict:
                    r1_ctg_ref_matched_pos_dict = current_read_base_r1_ctg_ref_dict[r1_ctg_ref]

                    # one read need to mapped to one ctg only for one time
                    if len(r1_ctg_ref_matched_pos_dict) > 1:
                        refs_to_ignore_ctg.add(r1_ctg_ref)
                    else:
                        r1_ctg_ref_pos = list(r1_ctg_ref_matched_pos_dict.keys())[0]
                        r1_ctg_ref_cigar = r1_ctg_ref_matched_pos_dict[r1_ctg_ref_pos]
                        r1_ctg_ref_len = ctg_len_dict[r1_ctg_ref]
                        qualified_cigar = check_cigar_quality(r1_ctg_ref_cigar, mismatch_cutoff, min_M_len_ctg,
                                                              r1_ctg_ref_pos, r1_ctg_ref_len)
                        if qualified_cigar is True:
                            r1_ctg_refs_passed_qc[r1_ctg_ref] = [r1_ctg_ref_cigar]
                        else:
                            refs_to_ignore_ctg.add(r1_ctg_ref)

                ########## filter r2 ctg refs ##########

                r2_ctg_refs_passed_qc = {}
                for r2_ctg_ref in current_read_base_r2_ctg_ref_dict:
                    r2_ctg_ref_matched_pos_dict = current_read_base_r2_ctg_ref_dict[r2_ctg_ref]

                    # one read need to mapped to one ctg only for one time
                    if len(r2_ctg_ref_matched_pos_dict) > 1:
                        refs_to_ignore_ctg.add(r2_ctg_ref)
                    else:
                        r2_ctg_ref_pos = list(r2_ctg_ref_matched_pos_dict.keys())[0]
                        r2_ctg_ref_cigar = r2_ctg_ref_matched_pos_dict[r2_ctg_ref_pos]
                        r2_ctg_ref_len = ctg_len_dict.get(r2_ctg_ref, 0)
                        qualified_cigar = check_cigar_quality(r2_ctg_ref_cigar, mismatch_cutoff, min_M_len_ctg,
                                                              r2_ctg_ref_pos, r2_ctg_ref_len)
                        if qualified_cigar is True:
                            r2_ctg_refs_passed_qc[r2_ctg_ref] = [r2_ctg_ref_cigar]
                        else:
                            refs_to_ignore_ctg.add(r2_ctg_ref)

                ####################################################################################################

                r1_ctg_refs_no_ignored = {key: value for key, value in r1_ctg_refs_passed_qc.items() if key not in refs_to_ignore_ctg}
                r2_ctg_refs_no_ignored = {key: value for key, value in r2_ctg_refs_passed_qc.items() if key not in refs_to_ignore_ctg}

                # no mate matched to ctg
                if (len(r1_ctg_refs_no_ignored) == 0) and (len(r2_ctg_refs_no_ignored) == 0):
                    pass

                # only r1 matched to ctg
                elif (len(r1_ctg_refs_no_ignored) > 0) and (len(r2_ctg_refs_no_ignored) == 0):

                    if current_read_base not in MappingRecord_dict:
                        MappingRecord_dict[current_read_base] = MappingRecord()

                    MappingRecord_dict[current_read_base].matched_to_ctg = True
                    MappingRecord_dict[current_read_base].r1_ctg_refs_no_ignored = r1_ctg_refs_no_ignored
                    MappingRecord_dict[current_read_base].r1_ctg_ref_dict = current_read_base_r1_ctg_ref_dict
                # only r2 matched to ctg
                elif (len(r1_ctg_refs_no_ignored) == 0) and (len(r2_ctg_refs_no_ignored) > 0):

                    if current_read_base not in MappingRecord_dict:
                        MappingRecord_dict[current_read_base] = MappingRecord()

                    MappingRecord_dict[current_read_base].matched_to_ctg = True
                    MappingRecord_dict[current_read_base].r2_ctg_refs_no_ignored = r2_ctg_refs_no_ignored
                    MappingRecord_dict[current_read_base].r2_ctg_ref_dict = current_read_base_r2_ctg_ref_dict

                # both r1 and r2 matched to ctg
                else:
                    shared_ctg_ref_set = {key: [r1_ctg_refs_no_ignored[key][0], r2_ctg_refs_no_ignored[key][0]] for key in set(r1_ctg_refs_no_ignored).intersection(set(r2_ctg_refs_no_ignored))}
                    if len(shared_ctg_ref_set) > 0:

                        if current_read_base not in MappingRecord_dict:
                            MappingRecord_dict[current_read_base] = MappingRecord()

                        MappingRecord_dict[current_read_base].matched_to_ctg = True
                        MappingRecord_dict[current_read_base].shared_ctg_refs_no_ignored = shared_ctg_ref_set
                        MappingRecord_dict[current_read_base].r1_ctg_ref_dict = current_read_base_r1_ctg_ref_dict
                        MappingRecord_dict[current_read_base].r2_ctg_ref_dict = current_read_base_r2_ctg_ref_dict

                ########################################### reset values ###########################################

                current_read_base = read_id_base
                current_read_base_r1_ctg_ref_dict = dict()
                current_read_base_r2_ctg_ref_dict = dict()

                if cigar != '*':
                    if read_strand == '1':
                        current_read_base_r1_ctg_ref_dict[ref_id] = {ref_pos: cigar}
                    if read_strand == '2':
                        current_read_base_r2_ctg_ref_dict[ref_id] = {ref_pos: cigar}

##################################################### get linkages from MappingRecord_dict #####################################################

print('%s %s' % ((datetime.now().strftime(time_format)), 'get linkages from MappingRecord_dict'))

marker_to_ctg_linkage_cigar_dict_16s_side = {}
marker_to_ctg_linkage_cigar_dict_ctg_side = {}
marker_to_ctg_linkage_num_dict = {}
marker_to_ctg_linking_reads_dict = {}
for qualified_read in MappingRecord_dict:

    r1_16s_refs     = MappingRecord_dict[qualified_read].r1_16s_refs_no_ignored
    r2_16s_refs     = MappingRecord_dict[qualified_read].r2_16s_refs_no_ignored
    shared_16s_refs = MappingRecord_dict[qualified_read].shared_16s_refs_no_ignored
    r1_ctg_refs     = MappingRecord_dict[qualified_read].r1_ctg_refs_no_ignored
    r2_ctg_refs     = MappingRecord_dict[qualified_read].r2_ctg_refs_no_ignored
    shared_ctg_refs = MappingRecord_dict[qualified_read].shared_ctg_refs_no_ignored

    # combine dict
    combined_16s_refs_dict = {**r1_16s_refs, **r2_16s_refs, **shared_16s_refs}
    combined_ctg_refs_dict = {**r1_ctg_refs, **r2_ctg_refs, **shared_ctg_refs}

    if (len(combined_16s_refs_dict) > 0) and (len(combined_ctg_refs_dict) > 0):
        for each_16s_ref in combined_16s_refs_dict:
            current_cigar_list_16s_side = combined_16s_refs_dict[each_16s_ref]
            for each_ctg_ref in combined_ctg_refs_dict:
                current_cigar_list_ctg_side = combined_ctg_refs_dict[each_ctg_ref]

                marker_to_ctg_key = '%s%s%s' % (each_16s_ref, marker_to_ctg_gnm_Key_connector, each_ctg_ref)

                # initialize key values if not in the dict
                if marker_to_ctg_key not in marker_to_ctg_linkage_num_dict:
                    marker_to_ctg_linkage_num_dict[marker_to_ctg_key] = 0
                    marker_to_ctg_linking_reads_dict[marker_to_ctg_key] = set()
                    marker_to_ctg_linkage_cigar_dict_16s_side[marker_to_ctg_key] = []
                    marker_to_ctg_linkage_cigar_dict_ctg_side[marker_to_ctg_key] = []

                # add values to dict
                marker_to_ctg_linkage_num_dict[marker_to_ctg_key] += 1
                marker_to_ctg_linking_reads_dict[marker_to_ctg_key].add(qualified_read)
                for each in current_cigar_list_16s_side:
                    marker_to_ctg_linkage_cigar_dict_16s_side[marker_to_ctg_key].append(each)
                for each in current_cigar_list_ctg_side:
                    marker_to_ctg_linkage_cigar_dict_ctg_side[marker_to_ctg_key].append(each)





print('\nFinally got:')
print(marker_to_ctg_linking_reads_dict['DA_m2___M___DA___C___NODE_537_length_39828_cov_427.956453'])
print(marker_to_ctg_linkage_num_dict['DA_m2___M___DA___C___NODE_537_length_39828_cov_427.956453'])
print(marker_to_ctg_linkage_cigar_dict_16s_side['DA_m2___M___DA___C___NODE_537_length_39828_cov_427.956453'])
print(marker_to_ctg_linkage_cigar_dict_ctg_side['DA_m2___M___DA___C___NODE_537_length_39828_cov_427.956453'])
print()





# calculate clp_pct_ratio
rd1_clp_pct_diff_txt_handle = open(rd1_clp_pct_diff_txt, 'w')
rd1_clp_pct_diff_txt_to_ignore_handle = open(rd1_clp_pct_diff_txt_to_ignore, 'w')
rd1_clp_pct_diff_txt_handle.write('Marker\tGenome\tContig\tclp_pct_ctg\tclp_pct_16s\tRatio\n')
rd1_clp_pct_diff_txt_to_ignore_handle.write('Marker\tGenome\tContig\tclp_pct_ctg\tclp_pct_16s\tRatio\n')
linkages_to_ignore = set()
for each_link in marker_to_ctg_linkage_num_dict:
    if marker_to_ctg_linkage_num_dict[each_link] >= 3:
        each_link_split = each_link.split(marker_to_ctg_gnm_Key_connector)
        marker_id = each_link_split[0]
        gnm_id = each_link_split[1].split(gnm_to_ctg_connector)[0]
        ctg_id = each_link_split[1].split(gnm_to_ctg_connector)[1]

        linkage_cigar_16s_side_all = marker_to_ctg_linkage_cigar_dict_16s_side[each_link]
        linkage_cigar_16s_side_clp = [i for i in linkage_cigar_16s_side_all if (('S' in i) or ('s' in i))]
        linkage_cigar_16s_side_clp_num = len(linkage_cigar_16s_side_clp)

        linkage_cigar_ctg_side_all = marker_to_ctg_linkage_cigar_dict_ctg_side[each_link]
        linkage_cigar_ctg_side_clp = [i for i in linkage_cigar_ctg_side_all if (('S' in i) or ('s' in i))]
        linkage_cigar_ctg_side_clp_num = len(linkage_cigar_ctg_side_clp)

        # if each_link == 'DA_m2___M___DA___C___NODE_537_length_39828_cov_427.956453':
        #     print('%s\tlinkage_cigar_16s_side_all: %s' % (each_link, linkage_cigar_16s_side_all))
        #     print('%s\tlinkage_cigar_16s_side_clp: %s' % (each_link, linkage_cigar_16s_side_clp))
        #     print('%s\tlinkage_cigar_ctg_side_all: %s' % (each_link, linkage_cigar_ctg_side_all))
        #     print('%s\tlinkage_cigar_ctg_side_clp: %s' % (each_link, linkage_cigar_ctg_side_clp))
        #     print()

        mean_clp_num = (linkage_cigar_16s_side_clp_num + linkage_cigar_ctg_side_clp_num)/2

        to_ignore = False
        if (linkage_cigar_16s_side_clp_num == 0) and (linkage_cigar_ctg_side_clp_num >= 10):
            to_ignore = True
        elif (linkage_cigar_16s_side_clp_num >= 10) and (linkage_cigar_ctg_side_clp_num == 0):
            to_ignore = True
        elif (linkage_cigar_16s_side_clp_num > 0) and (linkage_cigar_ctg_side_clp_num > 0):
            if mean_clp_num <= 20:
                if not (0.33 <= (linkage_cigar_16s_side_clp_num/linkage_cigar_ctg_side_clp_num) <= 3):
                    to_ignore = True
            # elif 20 < mean_clp_num <= 50:
            #     if not (0.5<= (linkage_cigar_16s_side_clp_num/linkage_cigar_ctg_side_clp_num) <= 2):
            #         to_ignore = True
            elif mean_clp_num > 20:
                if not (0.5 <= (linkage_cigar_16s_side_clp_num/linkage_cigar_ctg_side_clp_num) <= 2):
                    to_ignore = True

        # clp_pct_16s_side = 'na'
        # if len(linkage_cigar_16s_side_all) > 0:
        #     clp_pct_16s_side = len(linkage_cigar_16s_side_clp) * 100 / len(linkage_cigar_16s_side_all)
        #     clp_pct_16s_side = float("{0:.2f}".format(clp_pct_16s_side))
        #
        # clp_pct_ctg_side = 'na'
        # if len(linkage_cigar_ctg_side_all) > 0:
        #     clp_pct_ctg_side = len(linkage_cigar_ctg_side_clp) * 100 / len(linkage_cigar_ctg_side_all)
        #     clp_pct_ctg_side = float("{0:.2f}".format(clp_pct_ctg_side))
        #
        # clp_pct_ratio = 'na'
        # if (clp_pct_16s_side != 'na') and (clp_pct_ctg_side != 'na'):
        #     if clp_pct_16s_side > 0:
        #         clp_pct_ratio = float("{0:.2f}".format(clp_pct_ctg_side / clp_pct_16s_side))
        #
        # to_ignore = False
        # if clp_pct_ratio != 'na':
        #     if (clp_pct_ctg_side >= clp_pct_ctg_side_max_num) and (clp_pct_ratio >= clp_pct_ratio_cutoff):
        #         to_ignore = True
        # else:
        #     if (clp_pct_ctg_side != 'na') and (clp_pct_16s_side != 'na'):
        #         if clp_pct_ctg_side >= clp_pct_ctg_side_max_num:
        #             to_ignore = True

        if to_ignore is True:
            linkages_to_ignore.add(each_link)
            rd1_clp_pct_diff_txt_to_ignore_handle.write('%s\t%s\t%s\t%s/%s\t%s/%s\n'  % (marker_id, gnm_id, ctg_id, len(linkage_cigar_ctg_side_clp), len(linkage_cigar_ctg_side_all), len(linkage_cigar_16s_side_clp),len(linkage_cigar_16s_side_all)))
        else:
            rd1_clp_pct_diff_txt_handle.write('%s\t%s\t%s\t%s/%s\t%s/%s\n'            % (marker_id, gnm_id, ctg_id, len(linkage_cigar_ctg_side_clp), len(linkage_cigar_ctg_side_all), len(linkage_cigar_16s_side_clp), len(linkage_cigar_16s_side_all)))
rd1_clp_pct_diff_txt_handle.close()
rd1_clp_pct_diff_txt_to_ignore_handle.close()

# ignore linkages if did not pass the clp pct ratio test
marker_to_ctg_linkage_num_dict_min3_passed_ratio_check = {}
for each_link in marker_to_ctg_linkage_num_dict:
    if marker_to_ctg_linkage_num_dict[each_link] >= 3:
        if each_link not in linkages_to_ignore:
            marker_to_ctg_linkage_num_dict_min3_passed_ratio_check[each_link] = marker_to_ctg_linkage_num_dict[
                each_link]

# get number of linkages at genome level
marker_to_gnm_link_num = {}
for each_marker_to_ctg_key in marker_to_ctg_linkage_num_dict_min3_passed_ratio_check:
    marker_id = each_marker_to_ctg_key.split(marker_to_ctg_gnm_Key_connector)[0]
    ctg_id = each_marker_to_ctg_key.split(marker_to_ctg_gnm_Key_connector)[1]
    gnm_id = ctg_id.split(gnm_to_ctg_connector)[0]
    marker_to_gnm_key = '%s%s%s' % (marker_id, marker_to_ctg_gnm_Key_connector, gnm_id)
    if marker_to_gnm_key not in marker_to_gnm_link_num:
        marker_to_gnm_link_num[marker_to_gnm_key] = marker_to_ctg_linkage_num_dict_min3_passed_ratio_check[
            each_marker_to_ctg_key]
    else:
        marker_to_gnm_link_num[marker_to_gnm_key] += marker_to_ctg_linkage_num_dict_min3_passed_ratio_check[
            each_marker_to_ctg_key]

# write out linkages at genome level
sankey_file_in_handle = open(link_stats_combined, 'w')
sankey_file_in_handle.write('MarkerGene,GenomicSeq,Number\n')
for each_linkage in marker_to_gnm_link_num:
    sankey_file_in_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (
        each_linkage.split(marker_to_ctg_gnm_Key_connector)[0],
        each_linkage.split(marker_to_ctg_gnm_Key_connector)[1],
        marker_to_gnm_link_num[each_linkage]))
sankey_file_in_handle.close()

# write out linking reads
linking_reads_rd1_handle = open(linking_reads_rd1, 'w')
for each_link in marker_to_ctg_linkage_num_dict:
    current_linking_reads = marker_to_ctg_linking_reads_dict[each_link]
    linking_reads_rd1_handle.write('%s\t%s\n' % (each_link, ','.join(current_linking_reads)))
linking_reads_rd1_handle.close()


####################################### filter_linkages_iteratively ########################################

print('%s %s' % ((datetime.now().strftime(time_format)), 'filter_linkages_iteratively'))

pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, min_aln_16s, min_cov_16s)

filter_linkages_iteratively(link_stats_combined, 'Number', pairwise_16s_iden_dict,
                            mean_depth_dict_gnm, mean_depth_dict_16s, min_16s_gnm_multiple,
                            min_iden_16s, min_link_num,
                            min_link_num, within_gnm_linkage_num_diff, link_stats_combined_filtered_s1)




import os

def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


###################################################### file in/out #####################################################

wd                  = '/Users/songweizhi/Desktop/MarkerMAG_wd/4_Kelp'
taxonomy_mag        = '%s/BH_ER_050417.bac120.summary.tsv'                                              % wd
taxonomy_16s        = '%s/Kelp_SILVA138_id99_assembled_16S_uclust_0.999_classified.txt'                 % wd
taxonomy_16s        = '%s/Kelp_SILVA138_id99_assembled_16S_uclust_0.999_vs_GTDB_classifications.txt'    % wd
#taxonomy_16s        = '%s/Kelp_SILVA138_id99_assembled_16S_uclust_0.999_classified_corrected.txt'      % wd
taxonomy_16s_blca   = '%s/Kelp_SILVA138_id99_assembled_16S_uclust_0.999.fasta.blca.out'                 % wd
bin_id_file         = '%s/bin_id.txt'                                                                   % wd
barrnap_output      = '%s/BH_ER_050417_16S.txt'                                                         % wd
checkm_output       = '%s/BH_ER_050417_refined_MAG_qualities.txt'                                       % wd
mean_depth_file     = '%s/BH_ER_050417_refined_bins_mean_depth.txt'                                     % wd
gtdb_gnm_to_ssu_txt = '%s/GTDB_ssu_all_r95_gnm_to_ssu.txt'                                              % wd
# Kelp_0406_very_sensitive_identified_linkages_genome_level.txt
# Kelp_0406_sensitive_identified_linkages_genome_level.txt
# Kelp_0406_default_identified_linkages_genome_level.txt
# Kelp_0406_specific_identified_linkages_genome_level.txt
# Kelp_0406_very_specific_identified_linkages_genome_level.txt
# Kelp_0406_super_specific_identified_linkages_genome_level.txt
linkage_file = '/Users/songweizhi/Desktop/assess_linkages_Kelp/Kelp_0721_60_60_900_linkages_by_genome.txt'


'''

=============================
Linkage accuracy: Kelp_0721_60_60_900_linkages_by_genome
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	38	0	38	0	38/38(100.0)
class	38	0	36	0	36/38(94.74)
order	38	0	34	0	34/38(89.47)
family	38	0	32	0	32/38(84.21)
genus	38	0	19	8	27/38(71.05)
=============================
linked MAGs: 24


=============================
Linkage accuracy: Kelp_0721_45_45_900_linkages_by_genome
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	38	0	38	0	38/38(100.0)
class	38	0	36	0	36/38(94.74)
order	38	0	34	0	34/38(89.47)
family	38	0	32	0	32/38(84.21)
genus	38	0	20	7	27/38(71.05)
=============================
linked MAGs: 22



Linkage accuracy: Kelp_0602_60_60_linkages_by_genome
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	59	1	58	0	58/58(100.0)
class	59	1	56	0	56/58(96.55)
order	59	1	54	0	54/58(93.1)
family	59	1	49	0	49/58(84.48)
genus	59	1	32	12	44/58(75.86)

Linkage accuracy: Kelp_0413_very_sensitive_identified_linkages_genome_level
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	44	3	41	0	41/41(100.0)
class	44	3	40	0	40/41(97.56)
order	44	3	37	0	37/41(90.24)
family	44	3	34	0	34/41(82.93)
genus	44	3	23	6	29/41(70.73)
linked MAGs: 31

Linkage accuracy: Kelp_0413_sensitive_identified_linkages_genome_level
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	45	3	42	0	42/42(100.0)
class	45	3	41	0	41/42(97.62)
order	45	3	38	0	38/42(90.48)
family	45	3	35	0	35/42(83.33)
genus	45	3	22	8	30/42(71.43)
linked MAGs: 31

Linkage accuracy: Kelp_0413_default_identified_linkages_genome_level
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	39	2	37	0	37/37(100.0)
class	39	2	36	0	36/37(97.3)
order	39	2	33	0	33/37(89.19)
family	39	2	30	0	30/37(81.08)
genus	39	2	19	7	26/37(70.27)
linked MAGs: 29

Linkage accuracy: Kelp_0413_specific_identified_linkages_genome_level
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	35	4	31	0	31/31(100.0)
class	35	4	30	0	30/31(96.77)
order	35	4	27	0	27/31(87.1)
family	35	4	24	0	24/31(77.42)
genus	35	4	15	6	21/31(67.74)
linked MAGs: 25

Linkage accuracy: Kelp_0413_very_specific_identified_linkages_genome_level
Rank	All	NA	Correct	Unknown	Total(Accuracy)
phylum	33	2	31	0	31/31(100.0)
class	33	2	30	0	30/31(96.77)
order	33	2	27	0	27/31(87.1)
family	33	2	24	0	24/31(77.42)
genus	33	2	15	7	22/31(70.97)
linked MAGs: 24

'''


################################################### define file name ###################################################

linkage_file_path, linkage_file_basename, linkage_file_extension = sep_path_basename_ext(linkage_file)
linkage_file_with_assessment        = '%s/%s_with_assessment%s'         % (linkage_file_path, linkage_file_basename, linkage_file_extension)
linkage_file_with_assessment_by_mag = '%s/%s_with_assessment_by_MAG%s'  % (linkage_file_path, linkage_file_basename, linkage_file_extension)


################################################### read in metadata ###################################################

gtdb_gnm_to_ssu_dict = {}
for each_gnm in open(gtdb_gnm_to_ssu_txt):
    each_gnm_split = each_gnm.strip().split('\t')
    gnm_id = each_gnm_split[0]
    ssu_id_list = each_gnm_split[1].split(',')
    gtdb_gnm_to_ssu_dict[gnm_id] = ssu_id_list

bin_id_list = []
for each_bin_id in open(bin_id_file):
    bin_id_list.append(each_bin_id.strip())

bin_closest_ref_dict = {}
bin_taxon_dict = {}
for each_bin_taxon in open(taxonomy_mag):
    if not each_bin_taxon.startswith('user_genome	classification'):
        each_bin_taxon_split = each_bin_taxon.strip().split('\t')
        closest_ref = each_bin_taxon_split[7]
        bin_taxon_dict[each_bin_taxon_split[0]] = each_bin_taxon_split[1]
        bin_closest_ref_dict[each_bin_taxon_split[0]] = closest_ref

mag_completeness_dict = {}
for each_mag_quality in open(checkm_output):
    each_mag_quality_split = each_mag_quality.strip().split(',')
    mag_completeness_dict[each_mag_quality_split[0]] = each_mag_quality_split[1]

bin_depth_dict = {}
for each_bin_depth in open(mean_depth_file):
    each_bin_depth_split = each_bin_depth.strip().split('\t')
    bin_depth_dict[each_bin_depth_split[0]] = each_bin_depth_split[2]

barrnap_op_dict = {}
for each_barrnap_op in open(barrnap_output):
    if not each_barrnap_op.startswith('Genome	16S	Length'):
        each_barrnap_op_split = each_barrnap_op.strip().split('\t')
        barrnap_op_dict[each_barrnap_op_split[0]] = each_barrnap_op_split[2]

s16_taxon_dict = {}
s16_iden_dict = {}
s16_aln_dict = {}
for each_16s_taxon in open(taxonomy_16s):
    each_16s_taxon_split = each_16s_taxon.strip().split('\t')
    id_16s = each_16s_taxon_split[0]
    taxon_best_hit = each_16s_taxon_split[4].split(' [')[0]
    iden_with_best_match = each_16s_taxon_split[2]
    aln_with_best_match = each_16s_taxon_split[3]
    s16_taxon_dict[id_16s] = taxon_best_hit
    s16_iden_dict[id_16s]  = iden_with_best_match
    s16_aln_dict[id_16s]   = aln_with_best_match

s16_taxon_blca_dict = {}
for each_16s_taxon in open(taxonomy_16s_blca):
    each_16s_taxon_split = each_16s_taxon.strip().split('\t')
    s16_taxon_blca_dict[each_16s_taxon_split[0]] = each_16s_taxon_split[1]

s16_taxon_blca_dict_formatted = {}
for each_16s in s16_taxon_blca_dict:
    taxon_blca_raw = s16_taxon_blca_dict[each_16s]
    formatted_taxon_str = 'Unclassified'
    if taxon_blca_raw != 'Unclassified':
        taxon_blca_raw_split_1 = taxon_blca_raw.strip().split(':')[1:]
        formatted_taxon_list = []
        for each_str in taxon_blca_raw_split_1:
            each_str_split = each_str.split(';')
            taxon_with_confidence = '%s(%s)' % (each_str_split[0], each_str_split[1][:5])
            formatted_taxon_list.append(taxon_with_confidence)
        formatted_taxon_str = ';'.join(formatted_taxon_list)
    s16_taxon_blca_dict_formatted[each_16s] = formatted_taxon_str


###################################################### by linkage ######################################################

linkage_file_with_assessment_handle = open(linkage_file_with_assessment, 'w')
total_linkage = 0
correct_linkage_p = 0
linkage_with_unknown_p = 0
correct_linkage_c = 0
linkage_with_unknown_c = 0
correct_linkage_o = 0
linkage_with_unknown_o = 0
correct_linkage_f = 0
linkage_with_unknown_f = 0
correct_linkage_g = 0
linkage_with_unknown_g = 0
correct_linkage_s = 0
linkage_with_unknown_s = 0
unknown_mag_or_16s_taxon = 0
for each_linkage in open(linkage_file):
    if each_linkage.startswith('MarkerGene	GenomicSeq'):
        linkage_file_with_assessment_handle.write('%s\tPhylum\tClass\tOrder\tFamily\tGenus\tIdentity\tAln_len\tclosest_ref\tclosest_ref_16s_num\tMAG_taxon\t16S_taxon\t16S_taxon_BLCA\n' % each_linkage.strip())
    else:
        each_linkage_split = each_linkage.strip().split('\t')
        line_to_write = each_linkage.strip()
        s16_id = each_linkage_split[0]
        mag_id = each_linkage_split[1]
        taxon_mag = bin_taxon_dict.get(mag_id, 'NA')
        closest_ref = bin_closest_ref_dict.get(mag_id, 'NA')
        closest_ref_16s_num = len(gtdb_gnm_to_ssu_dict.get(closest_ref, []))
        taxon_16s = s16_taxon_dict.get(s16_id, 'NA')
        iden_16s = s16_iden_dict.get(s16_id, 'NA')
        aln_len_16s = s16_aln_dict.get(s16_id, 'NA')
        taxon_16s_blca = s16_taxon_blca_dict_formatted.get(s16_id, 'NA')
        taxon_mag_split = taxon_mag.split(';')
        taxon_16s_split = taxon_16s.split(';')
        total_linkage += 1

        if (taxon_mag == 'NA') or (taxon_16s == 'NA'):
            unknown_mag_or_16s_taxon += 1
            line_to_write += '\tna\tna\tna\tna\tna'
        else:
            taxon_mag_p = taxon_mag_split[1]
            taxon_mag_c = taxon_mag_split[2]
            taxon_mag_o = taxon_mag_split[3]
            taxon_mag_f = taxon_mag_split[4]
            taxon_mag_g = taxon_mag_split[5]
            taxon_mag_s = taxon_mag_split[6]
            taxon_16s_p = taxon_16s_split[1]
            taxon_16s_c = taxon_16s_split[2]
            taxon_16s_o = taxon_16s_split[3]
            taxon_16s_f = taxon_16s_split[4]
            taxon_16s_g = taxon_16s_split[5]
            taxon_16s_s = taxon_16s_split[6]
            already_wrong = False

            if taxon_mag_p == taxon_16s_p:
                correct_linkage_p += 1
                line_to_write += '\t1'
            else:
                if taxon_mag_p == 'p__':
                    linkage_with_unknown_p += 1
                    line_to_write += '\tna'
                else:
                    already_wrong = True
                    line_to_write += '\t0'

            if already_wrong is False:
                if taxon_mag_c == taxon_16s_c:
                    correct_linkage_c += 1
                    line_to_write += '\t1'
                else:
                    if taxon_mag_c == 'c__':
                        linkage_with_unknown_c += 1
                        line_to_write += '\tna'
                    else:
                        already_wrong = True
                        line_to_write += '\t0'
            else:
                line_to_write += '\t0'

            if already_wrong is False:
                if taxon_mag_o == taxon_16s_o:
                    correct_linkage_o += 1
                    line_to_write += '\t1'
                else:
                    if taxon_mag_o == 'o__':
                        linkage_with_unknown_o += 1
                        line_to_write += '\tna'
                    else:
                        already_wrong = True
                        line_to_write += '\t0'
            else:
                line_to_write += '\t0'

            if already_wrong is False:
                if taxon_mag_f == taxon_16s_f:
                    correct_linkage_f += 1
                    line_to_write += '\t1'
                else:
                    if taxon_mag_f == 'f__':
                        linkage_with_unknown_f += 1
                        line_to_write += '\tna'
                    else:
                        already_wrong = True
                        line_to_write += '\t0'
            else:
                line_to_write += '\t0'

            if already_wrong is False:
                if taxon_mag_g == taxon_16s_g:
                    correct_linkage_g += 1
                    line_to_write += '\t1'
                else:
                    if taxon_mag_g == 'g__':
                        linkage_with_unknown_g += 1
                        line_to_write += '\tna'
                    else:
                        already_wrong = True
                        line_to_write += '\t0'
            else:
                line_to_write += '\t0'

        line_to_write += '\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % (iden_16s, aln_len_16s, closest_ref, closest_ref_16s_num, taxon_mag, taxon_16s, taxon_16s_blca)
        linkage_file_with_assessment_handle.write('%s\n' % line_to_write)


if total_linkage > 0:

    print(total_linkage)
    print()
    print('=============================')
    accuracy_p = float("{0:.2f}".format((correct_linkage_p + linkage_with_unknown_p) * 100 / (total_linkage - unknown_mag_or_16s_taxon)))
    accuracy_c = float("{0:.2f}".format((correct_linkage_c + linkage_with_unknown_c) * 100 / (total_linkage - unknown_mag_or_16s_taxon)))
    accuracy_o = float("{0:.2f}".format((correct_linkage_o + linkage_with_unknown_o) * 100 / (total_linkage - unknown_mag_or_16s_taxon)))
    accuracy_f = float("{0:.2f}".format((correct_linkage_f + linkage_with_unknown_f) * 100 / (total_linkage - unknown_mag_or_16s_taxon)))
    accuracy_g = float("{0:.2f}".format((correct_linkage_g + linkage_with_unknown_g) * 100 / (total_linkage - unknown_mag_or_16s_taxon)))

    print('Linkage accuracy: %s' % linkage_file_basename)
    print('Rank\tAll\tNA\tCorrect\tUnknown\tTotal(Accuracy)')
    print('phylum\t%s\t%s\t%s\t%s\t%s/%s(%s)' % (total_linkage, unknown_mag_or_16s_taxon, correct_linkage_p, linkage_with_unknown_p, (correct_linkage_p + linkage_with_unknown_p), (total_linkage - unknown_mag_or_16s_taxon), accuracy_p))
    print('class\t%s\t%s\t%s\t%s\t%s/%s(%s)'  % (total_linkage, unknown_mag_or_16s_taxon, correct_linkage_c, linkage_with_unknown_c, (correct_linkage_c + linkage_with_unknown_c), (total_linkage - unknown_mag_or_16s_taxon), accuracy_c))
    print('order\t%s\t%s\t%s\t%s\t%s/%s(%s)'  % (total_linkage, unknown_mag_or_16s_taxon, correct_linkage_o, linkage_with_unknown_o, (correct_linkage_o + linkage_with_unknown_o), (total_linkage - unknown_mag_or_16s_taxon), accuracy_o))
    print('family\t%s\t%s\t%s\t%s\t%s/%s(%s)' % (total_linkage, unknown_mag_or_16s_taxon, correct_linkage_f, linkage_with_unknown_f, (correct_linkage_f + linkage_with_unknown_f), (total_linkage - unknown_mag_or_16s_taxon), accuracy_f))
    print('genus\t%s\t%s\t%s\t%s\t%s/%s(%s)'  % (total_linkage, unknown_mag_or_16s_taxon, correct_linkage_g, linkage_with_unknown_g, (correct_linkage_g + linkage_with_unknown_g), (total_linkage - unknown_mag_or_16s_taxon), accuracy_g))
    print('=============================')

######################################################## by MAG ########################################################

linked_mag_list = set()
for each_linkage in open(linkage_file):
    if not each_linkage.startswith('MarkerGene	GenomicSeq	Linkage	Step'):
        each_linkage_split = each_linkage.strip().split('\t')
        mag_id = each_linkage_split[1]
        linked_mag_list.add(mag_id)

linkage_file_with_assessment_by_mag_handle = open(linkage_file_with_assessment_by_mag, 'w')
linkage_file_with_assessment_by_mag_handle.write('MAG\tdepth\t16S\tcompleteness\tlinked\ttaxon\n')
for each_mag in sorted(bin_id_list):

    mag_depth = '0'
    if each_mag in bin_depth_dict:
        mag_depth = bin_depth_dict[each_mag]

    mag_taxon = 'unknown'
    if each_mag in bin_taxon_dict:
        mag_taxon = bin_taxon_dict[each_mag]

    with_16s_in_mag = 'no'
    if each_mag in barrnap_op_dict:
        with_16s_in_mag = barrnap_op_dict[each_mag]

    linked_to_16s = 'no'
    if each_mag in linked_mag_list:
        linked_to_16s = 'yes'

    mag_cpl = 0
    if each_mag in mag_completeness_dict:
        mag_cpl = mag_completeness_dict[each_mag]

    linkage_file_with_assessment_by_mag_handle.write('%s\t%s\t%s\t%s\t%s\t%s\n' % (each_mag, mag_depth, with_16s_in_mag, mag_cpl, linked_to_16s, mag_taxon))
linkage_file_with_assessment_by_mag_handle.close()

print('linked MAGs: %s' % len(linked_mag_list))

'''

Kelp_SILVA138_id99_subsample_75_306		Refined_20	42	S1	1	0	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Pseudomonadales;f__HTCC2089;g__SYFY01;s__	d__Bacteria;p__Proteobacteria;c__Alphaproteobacteria;o__Sphingomonadales;f__Sphingomonadaceae;g__Erythrobacter_A;s__Erythrobacter_A_sp002895025	Bacteria(100.0);Proteobacteria(100.0);Alphaproteobacteria(88.5);Sphingomonadales(88.5);Sphingomonadaceae(88.5);Erythrobacter_A(88.5);Erythrobacter_A sp002895025(88.5)
Kelp_SILVA138_id99_subsample_100_365	Refined_20	23	S1	1	0	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Pseudomonadales;f__HTCC2089;g__SYFY01;s__	d__Bacteria;p__Proteobacteria;c__Alphaproteobacteria;o__Sphingomonadales;f__Sphingomonadaceae;g__Erythrobacter_A;s__Erythrobacter_A_sp002895025	Bacteria(100.0);Proteobacteria(100.0);Alphaproteobacteria(85.0);Sphingomonadales(85.0);Sphingomonadaceae(85.0);Erythrobacter_A(85.0);Erythrobacter_A sp002895025(85.0)
Kelp_SILVA138_id99_subsample_25_212		Refined_20	11	S1	1	1	1	1	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Pseudomonadales;f__HTCC2089;g__SYFY01;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Pseudomonadales;f__HTCC2089;g__SZUA-521;s__SZUA-521_sp003248125	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(55.50);Pseudomonadales(55.50);HTCC2089(55.50);SZUA-521(55.50);SZUA-521 sp003248125(55.50)

Kelp_SILVA138_id99_subsample_75_668		Refined_26	91	S1	1	1	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__GCA-1730015;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Chromatiales;f__Sedimenticolaceae;g__QGON01;s__QGON01_sp003660235	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Chromatiales(81.33);Sedimenticolaceae(81.33);QGON01(81.33);QGON01 sp003660235(80.99)
Kelp_SILVA138_id99_subsample_50_553		Refined_26	20	S1	1	1	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__GCA-1730015;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Chromatiales;f__Sedimenticolaceae;g__QGON01;s__QGON01_sp003660235	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Granulosicoccales(78.33);Granulosicoccaceae(78.33);Granulosicoccus(78.33);Granulosicoccus antarcticus(78.33)
Kelp_SILVA138_id99_subsample_10_229		Refined_26	42	S1	1	1	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__GCA-1730015;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Chromatiales;f__Sedimenticolaceae;g__QGON01;s__QGON01_sp003660235	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Chromatiales(95.75);Sedimenticolaceae(95.75);QGON01(95.75);QGON01 sp003660235(91.50)
Kelp_SILVA138_id99_subsample_75_314		Refined_26	17	S1	1	1	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__GCA-1730015;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Chromatiales;f__Sedimenticolaceae;g__QGON01;s__QGON01_sp003660235	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Chromatiales(82.0);Sedimenticolaceae(82.0);QGON01(82.0);QGON01 sp003660235(45.0)
Kelp_SILVA138_id99_subsample_25_393		Refined_26	9	S1	1	1	0	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__GCA-1730015;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Chromatiales;f__Sedimenticolaceae;g__QGON01;s__QGON01_sp003660235	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Granulosicoccales(51.0);Granulosicoccaceae(51.0);Granulosicoccus(51.0);Granulosicoccus antarcticus(51.0)
Kelp_SILVA138_id99_subsample_100_789	Refined_26	13	S1	1	1	1	1	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__GCA-1730015;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Granulosicoccales;f__Granulosicoccaceae;g__Granulosicoccus;s__Granulosicoccus_antarcticus	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Granulosicoccales(97.0);Granulosicoccaceae(97.0);Granulosicoccus(97.0);Granulosicoccus antarcticus(97.0)

Kelp_SILVA138_id99_subsample_10_136		Refined_57	154	S1	1	1	1	0	0	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Pseudomonadales;f__Porticoccaceae;g__;s__	d__Bacteria;p__Proteobacteria;c__Gammaproteobacteria;o__Pseudomonadales;f__Cellvibrionaceae;g__Teredinibacter;s__Teredinibacter_sp000966245	Bacteria(100.0);Proteobacteria(100.0);Gammaproteobacteria(100.0);Pseudomonadales(100.0);Cellvibrionaceae(78.0);Teredinibacter(60.0);Teredinibacter sp003634075(60.0)

Kelp_SILVA138_id99_subsample_50_115		Refined_19	44	S1	1	1	1	0	0	d__Bacteria;p__Planctomycetota;c__UBA1135;o__UBA1135;f__GCA-002686595;g__;s__	d__Bacteria;p__Planctomycetota;c__UBA1135;o__UBA1135;f__UBA1135;g__GCA-2705055;s__GCA-2705055_sp002705055	Unclassified

Kelp_SILVA138_id99_subsample_50_418		Refined_31	28	S2	1	1	1	0	0	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__SHLQ01;g__SHLQ01;s__	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__MedAcidi-G1;g__S20-B6;s__S20-B6_sp002346745	Bacteria(100.0);Actinobacteriota(100.0);Acidimicrobiia(100.0);Acidimicrobiales(100.0);MedAcidi-G1(100.0);MedAcidi-G1(60.5);MedAcidi-G1 sp002713545(60.5)
Kelp_SILVA138_id99_subsample_75_575		Refined_31	28	S2	1	1	1	0	0	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__SHLQ01;g__SHLQ01;s__	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__MedAcidi-G1;g__S20-B6;s__S20-B6_sp002699725	Bacteria(100.0);Actinobacteriota(100.0);Acidimicrobiia(100.0);Acidimicrobiales(100.0);MedAcidi-G1(100.0);UBA3125(78.33);UBA3125 sp002687745(78.33)
Kelp_SILVA138_id99_subsample_100_694	Refined_31	28	S2	1	1	1	0	0	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__SHLQ01;g__SHLQ01;s__	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__MedAcidi-G1;g__S20-B6;s__S20-B6_sp002346745	Bacteria(100.0);Actinobacteriota(100.0);Acidimicrobiia(100.0);Acidimicrobiales(100.0);MedAcidi-G1(87.0);MedAcidi-G1(81.0);MedAcidi-G1 sp003214465(81.0)
Kelp_SILVA138_id99_subsample_75_493		Refined_31	3	S2	1	1	1	0	0	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__SHLQ01;g__SHLQ01;s__	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__MedAcidi-G1;g__MedAcidi-G1;s__MedAcidi-G1_sp002713545	Bacteria(100.0);Actinobacteriota(100.0);Acidimicrobiia(100.0);Acidimicrobiales(100.0);MedAcidi-G1(100.0);MedAcidi-G1(72.33);MedAcidi-G1 sp002713545(72.33)
Kelp_SILVA138_id99_subsample_50_487		Refined_31	8	S2	1	1	1	0	0	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__SHLQ01;g__SHLQ01;s__	d__Bacteria;p__Actinobacteriota;c__Acidimicrobiia;o__Acidimicrobiales;f__MedAcidi-G1;g__S20-B6;s__S20-B6_sp002346745	Bacteria(100.0);Actinobacteriota(100.0);Acidimicrobiia(100.0);Acidimicrobiales(100.0);MedAcidi-G1(57.5);MedAcidi-G3(42.5);MedAcidi-G3 sp000817105(42.5)


'''



depth_file_mag = '/Users/songweizhi/Desktop/666/hc_0513_with_depth_global_mean_depth_gnm.txt'
depth_file_16s = '/Users/songweizhi/Desktop/666/hc_0513_with_depth_global_mean_depth_16s.txt'
linkage_file   = '/Users/songweizhi/Desktop/666/hc_0514_stats_combined_filtered_with_assessment_ani97_imag99.5_i16S99.5.txt'


mean_depth_dict_gnm = {}
for each_mag_depth in open(depth_file_mag):
    each_mag_depth_split = each_mag_depth.strip().split('\t')
    mag_id = each_mag_depth_split[0]
    mag_depth = float(each_mag_depth_split[1])
    mean_depth_dict_gnm[mag_id] = mag_depth

mean_depth_dict_16s = {}
for each_16s_depth in open(depth_file_16s):
    each_16s_depth_split = each_16s_depth.strip().split('\t')
    s16_id = each_16s_depth_split[0]
    s16_depth = float(each_16s_depth_split[1])
    mean_depth_dict_16s[s16_id] = s16_depth

for each_link in open(linkage_file):
    each_link_split = each_link.strip().split('\t')
    id_16s = each_link_split[0]
    id_mag = each_link_split[1]

    depth_16s = mean_depth_dict_16s[id_16s]
    depth_mag = mean_depth_dict_gnm[id_mag]
    ratio = depth_16s/depth_mag
    depth_16s = float("{0:.2f}".format(depth_16s))
    depth_mag = float("{0:.2f}".format(depth_mag))
    ratio = float("{0:.2f}".format(ratio))

    print('%s\t%s\t%s\t%s' % (each_link.strip(), depth_16s, depth_mag, ratio))







sub_num_dict = {}
for each_line in open('/Users/songweizhi/Desktop/STY_Merged_OTU01-Transporter.tbl'):
    sub_name = each_line.strip().split('\t')[2]
    if sub_name not in sub_num_dict:
        sub_num_dict[sub_name] = 1
    else:
        sub_num_dict[sub_name] += 1

for each_sub in sub_num_dict:
    print('%s\t%s' % (each_sub, sub_num_dict[each_sub]))



import os
import glob
from Bio import SeqIO


genome_folder       = '/Users/songweizhi/Desktop/BH_ER_050417_refined_bins'
genome_ext          = 'fasta'
genome_folder_out   = '/Users/songweizhi/Desktop/BH_ER_050417_refined_bins_noNs'


file_re = '%s/*.%s' % (genome_folder, genome_ext)
file_list = [os.path.basename(file_name) for file_name in glob.glob(file_re)]


for genome in file_list:
    pwd_genome = '%s/%s' % (genome_folder, genome)
    pwd_genome_out = '%s/%s' % (genome_folder_out, genome)
    pwd_genome_out_tmp = '%s/%s.tmp' % (genome_folder_out, genome)
    pwd_genome_out_tmp_handle = open(pwd_genome_out_tmp, 'w')
    for seq_record in SeqIO.parse(pwd_genome, 'fasta'):
        pwd_genome_out_tmp_handle.write('>%s\n' % seq_record.id)
        pwd_genome_out_tmp_handle.write('%s\n' % seq_record.seq)
    pwd_genome_out_tmp_handle.close()


    split_cmd = 'perl /Users/songweizhi/Desktop/split.scaffolds.to.contigs.pl -i %s -o %s -m 2000' % (pwd_genome_out_tmp, pwd_genome_out)
    os.system(split_cmd)

    os.system('rm %s' % pwd_genome_out_tmp)






def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, within_gnm_linkage_num_diff, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:
                if MarkerGene not in MarkerGene_with_assignment:

                    # consider depth
                    if min_16s_gnm_multiple > 0:

                        # get marker and genome depth
                        MarkerGene_depth = marker_gene_depth_dict.get(MarkerGene, 'na')
                        GenomicSeq_depth = genomic_seq_depth_dict.get(GenomicSeq, 'na')
                        marker_genome_depth_ratio = 'na'
                        if (MarkerGene_depth != 'na') and (GenomicSeq_depth != 'na'):
                            if GenomicSeq_depth > 0:
                                marker_genome_depth_ratio = MarkerGene_depth / GenomicSeq_depth

                        if marker_genome_depth_ratio >= min_16s_gnm_multiple:
                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                gnm_max_link_num_dict[GenomicSeq] = linkage_num
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                # get identity with best marker
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                                iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                    if (linkage_num * 100 / gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                        file_out_handle.write(each_match)
                                        MarkerGene_with_assignment.add(MarkerGene)
                                    else:
                                        MarkerGene_with_assignment.add(MarkerGene)
                    # ignore depth
                    else:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            gnm_max_link_num_dict[GenomicSeq] = linkage_num
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            # get identity with best marker
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                            iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                if (linkage_num*100/gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                                else:
                                    MarkerGene_with_assignment.add(MarkerGene)
    file_out_handle.close()


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


#################################################################################################################

# file in
wd = '/Users/songweizhi/Desktop/filter'
link_stats_combined             = '%s/.txt' % wd
depth_file_mag                  = '%s/hc_0513_with_depth_global_mean_depth_gnm.txt' % wd
depth_file_16s                  = '%s/hc_0513_with_depth_global_mean_depth_16s.txt' % wd
blast_results_all_vs_all_16s    = ''
min_aln_16s                     = 500
min_cov_16s                     = 30
min_16s_gnm_multiple            = 0.3
min_iden_16s                    = 98
min_link_num                    = 8
within_gnm_linkage_num_diff     = 80

# file out
link_stats_combined_filtered_s1 = '%s/.txt' % wd


#################################################################################################################

mean_depth_dict_gnm = {}
for each_mag_depth in open(depth_file_mag):
    each_mag_depth_split = each_mag_depth.strip().split('\t')
    mag_id = each_mag_depth_split[0]
    mag_depth = float(each_mag_depth_split[1])
    mean_depth_dict_gnm[mag_id] = mag_depth

mean_depth_dict_16s = {}
for each_16s_depth in open(depth_file_16s):
    each_16s_depth_split = each_16s_depth.strip().split('\t')
    s16_id = each_16s_depth_split[0]
    s16_depth = float(each_16s_depth_split[1])
    mean_depth_dict_16s[s16_id] = s16_depth


pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, min_aln_16s, min_cov_16s)


####################################### filter_linkages_iteratively ########################################

filter_linkages_iteratively(link_stats_combined, 'Number', pairwise_16s_iden_dict,
                            mean_depth_dict_gnm, mean_depth_dict_16s, min_16s_gnm_multiple,
                            min_iden_16s, min_link_num,
                            min_link_num, within_gnm_linkage_num_diff, link_stats_combined_filtered_s1)










def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


input_reads_to_16s_sam_best_match = '/Users/songweizhi/Desktop/000000/ref_16s.sam'
min_M_pct = 35
mismatch_cutoff = 2

marker_len_dict = {}
for each_read in open(input_reads_to_16s_sam_best_match):
    each_read_split = each_read.strip().split('\t')

    if each_read.startswith('@'):
        marker_id = ''
        marker_len = 0
        for each_element in each_read_split:
            if each_element.startswith('SN:'):
                marker_id = each_element[3:]
            if each_element.startswith('LN:'):
                marker_len = int(each_element[3:])
        marker_len_dict[marker_id] = marker_len
    else:
        store_read_seq = False
        read_id = each_read_split[0]
        read_id_base = '.'.join(read_id.split('.')[:-1])
        read_strand = read_id.split('.')[-1]
        read_flag = int(each_read_split[1])
        read_seq = each_read_split[9]
        cigar = each_read_split[5]

        if cigar != '*':
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])
            ref_id_with_pos = '%s_pos_%s' % (ref_id, ref_pos)
            cigar_splitted = cigar_splitter(cigar)
            aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitted)
            both_ends_clp = check_both_ends_clipping(cigar_splitted)

            if (aligned_pct >= min_M_pct) and (mismatch_pct <= mismatch_cutoff) and (both_ends_clp is False):

                clip_in_middle = False
                if ('S' in cigar) or ('s' in cigar):
                    clip_in_middle = True
                    if (cigar_splitted[0][-1] in ['S', 's']) and (ref_pos == 1):
                        clip_in_middle = False
                    if (cigar_splitted[-1][-1] in ['S', 's']):
                        if (ref_pos + aligned_len - 1) == marker_len_dict[ref_id]:
                            clip_in_middle = False

                if clip_in_middle is False:
                    print(each_read_split)






